#ifndef CLUSTER_HH
#define CLUSTER_HH

#include "auxiliary.hh"
#include "functions.hh"

/* A cluster is a collection of variables and functions (incl. hard constraints)
   A cluster is a node of the Cluster-Tree-Decomposition
   Clusters are connected to their children and their parent
*/

class Cluster {
protected:
    std::string name_;
    var_ptrs_t vars_;
    func_ptrs_t funcs_;
    func_ptrs_t cons_;

    Cluster *parent_;
    cluster_ptrs_t children_;

    bool contains_var(Variable *v);
    int sum_funcs();
public:
    Cluster(const std::string &name): name_(name),parent_(NULL) {}; 
    
    virtual ~Cluster();

    void ask_children(); // ask the children for their messages
                         // the children answer by asking their own children
                         // and then sending their own message to the parent
    virtual MaterializedFunction *marginalize();
    void marginalize_on(var_ptrs_t &sep_vars,uint sep_vars_idx,var_ptrs_t &elim_vars,MaterializedFunction *f);
    int maximize_over(var_ptrs_t &elim_vars, uint elim_vars_idx, int maximum, std::vector<int> *pbest=0);

    void register_func(Function *f) {funcs_.push_back(f);}
    void register_cons(Function *f) {cons_.push_back(f);}
    void register_var(Variable *v) {vars_.push_back(v);}
    void register_child(Cluster *c) {children_.push_back(c); c->parent_=this;}

    void partition_shared_vars(var_ptrs_t &sep_vars,var_ptrs_t &elim_vars);

    bool propagate_to(Variable *v); // call all functions to propagate to variable v

    virtual void print_debug();
    void print_debug_rec();

    void free_rec();

    void backtrace();
};

#endif
