/************************************************************
 *
 * Alignment by Cluster Tree Elimination
 *
 * A rather general framework for CTE is used
 * for efficient constraint-based alignment
 *
 * a special case of constraint alignment with "segment constraints"
 * is implemented as extension to pure sequence alignment
 *
 ************************************************************/

//************************************************************
// There is still a bug in computation of score!!!
//************************************************************


#include "auxiliary.hh"
#include "functions.hh"
#include "cluster.hh"
#include "ctd.hh"

#include <fstream>
#include <iostream>

#include "leq_constraint.hh"

#include <math.h>

using namespace std;

#define BUF_SIZE 1024

// use specialized marginalization that is more efficient,
// than the pure CTE marginalization (whenever possible)
const int OPTIMIZED_ALIGNMENT = true;  // false;

/************************************************************
 * specialized functions and hard constraints
 * for sequence alignment
 */

/*
  Parameter/data for Scoring
*/
class ScoringInfo {
    string seqA;
    string seqB;
 
    void init_sequences(const string _seqA, const string _seqB) {
	seqA="";
	seqB="";
	for (int i=0; i<(int)_seqA.size(); i++)
	    seqA+=toupper(_seqA[i]);
	for (int i=0; i<(int)_seqB.size(); i++)
	    seqB+=toupper(_seqB[i]);
    }


public:
    const int gamma;
    //const int match;
    //const int mismatch;
    
    vector<vector<int> > sigma_tab;
    
    int sigma(uint i, uint j) const {
	if (i==seqA.size()+1 && j==seqB.size()+1) return 0; //special case 
	//cout <<"Sigma: "<<i-1<<" "<<j-1<<" "<<seqA[i-1]<<" "<<seqB[j-1]<<endl; 
	// return (seqA[i-1]==seqB[j-1])?match:mismatch;
	return sigma_tab[(unsigned char)seqA[i-1]][(unsigned char)seqB[j-1]];
    }
    
    // construction with match/mismatch cost
    ScoringInfo(const string _seqA, const string _seqB,
		int _gamma, int _match, int _mismatch)
	: gamma(_gamma) {
	
	init_sequences(_seqA,_seqB);
	
	sigma_tab.resize(256);
	for (int i=0; i<256; i++) {
	    sigma_tab[i].resize(256);
	    for (int j=0; j<256; j++) {
		sigma_tab[i][j] = _mismatch;
	    }
	}
	for (int i=0; i<256; i++) {
	    sigma_tab[i][i] = _match;
	}
    }
    
    // construction with similarity matrix from file
    ScoringInfo(const string _seqA, const string _seqB,
		int _gamma, string table)
	: gamma(_gamma) {
	
	init_sequences(_seqA,_seqB);	
	
	// read table
	// ...
	
	ifstream in(table.c_str());
	if (!in) {
	  cerr << "Cannot read file "<<table<<endl;
	  exit(-1);
	}
	read_sim_matrix(in);
    }
  
    void read_sim_matrix(istream &in);
    
};

// read similarity matrix from stream in
void ScoringInfo::read_sim_matrix(istream &in) {
  vector<unsigned char> alphabet;
  char line[BUF_SIZE];
  unsigned int lineno=0;
  
  sigma_tab.resize(256);
  for (unsigned int i=0; i<256; i++) {
    sigma_tab[i].resize(256);
    for (unsigned int j=0; j<256; j++) {
      sigma_tab[i][j]=0;
    }
  }
  
  int table_row=-1; // current table row

  while(in.getline(line,BUF_SIZE)) {
    lineno++;
    if (line[0]==0) continue;
    if (line[0]=='#') continue;
    istringstream lstr(line); // the line as a string stream
    
    if (line[0]==' ') {// header of table
      if (table_row != -1) { // error in file format
	cerr << "Wrong similarity matrix file format at line "
	     <<lineno<<" (unexpected table header)"<<endl;
	exit(-1);
      }
      string token;
      while(lstr >> token) {
	alphabet.push_back((unsigned char)token[0]);
      }
      table_row++;
    } else { // row of table
      string token;
      if (table_row>=0
	  && lstr >> token 
	  && (unsigned int)table_row < alphabet.size()
	  && token[0]==alphabet[table_row]) {
	// read <table_row>th row of the table for character alphabet[table_row]
	int score;
	unsigned int table_col=0;
	while (lstr >> score) {
	  sigma_tab[alphabet[table_row]][alphabet[table_col]]=score;
	  table_col ++;
	}
	table_row++;
      }
      else {
	cerr << "Wrong similarity matrix file format at line "
	     <<lineno<<", table row "<<table_row<<endl;
	exit(-1);
      }
    }
  }
  /* write table to stdout (for debugging)
  cout<<"Table:"<<endl;
  for (unsigned int i=0; i<alphabet.size(); i++) {
    cout<<alphabet[i]<<" ";
    for (unsigned int j=0; j<alphabet.size(); j++)
      cout << sigma_tab[alphabet[i]][alphabet[j]]<<" ";
    cout<<endl;
  }
  */
}

/*
  Function for positionwise score contributions
*/
class ScoreFunction : public Function {
private:
  const ScoringInfo *sc_;
  int i_;
public:
  ScoreFunction(const ScoringInfo *sc, int i, Variable *v0, Variable *v1);
  virtual ~ScoreFunction();

  virtual int value() const; // implements value method
};

ScoreFunction::~ScoreFunction() {}

ScoreFunction::ScoreFunction(const ScoringInfo *sc,
			     int i,
			     Variable *v0,
			     Variable *v1)
  : Function("f_"+itoa(i)),sc_(sc),i_(i)
{
  vars_.push_back(v0);
  vars_.push_back(v1);
}

int ScoreFunction::value() const {
  int j=vars_[1]->get_singleton();
  int j_pr=vars_[0]->get_singleton();
    
  if (j_pr < j) 
    return sc_->sigma(i_,j) + (j - j_pr -1)*sc_->gamma;
  else
    return sc_->gamma;
}

/************************************************************
 *
 * Cluster with special more efficient marginalization

   There is no clean interface to do such modifications yet
   ==> THIS IS A HACK !!!

*/

class AligCluster: public Cluster {
  const int i_;
  const ScoringInfo *sc_;
  bool fast_marg_; // can we do a fast, specialized
  // alignment marginalization
public:
  AligCluster(const string &name, int i, const ScoringInfo *sc)
    :Cluster("Alig-"+name),
     i_(i), 
     sc_(sc),
     fast_marg_(OPTIMIZED_ALIGNMENT) {}; 
  
  MaterializedFunction *marginalize();
  enum {GENERAL_MARG, FAST_MARG};
  void marginalization_method(int type);
    
    virtual void print_debug();
};


void AligCluster::marginalization_method(int type) {
  fast_marg_ = (type==FAST_MARG);
}

void AligCluster::print_debug() {
    Cluster::print_debug();
    cout << "Fast marg. "<<(fast_marg_?"on":"off")<<endl;
}


inline
int infty_add(int x, int y) {
  return (x==INT_MIN)?INT_MIN:(x+y);
}


MaterializedFunction *AligCluster::marginalize() {
    if ( ! fast_marg_ ) {
        return Cluster::marginalize();
    }
    
  ask_children();

  if (VERBOSE) cout <<"Fast marginalization in Cluster "<<name_<<endl;

  var_ptrs_t sep_vars;
  var_ptrs_t elim_vars;
  partition_shared_vars(sep_vars,elim_vars);
    
  MaterializedFunction *f = new MaterializedFunction("g_{"+name_+"}",sep_vars);
    
  Variable *sep_var=sep_vars[0];
  // Variable *elim_var=elim_vars[0];
    
  Domain dom=sep_var->get_dom();
  if (propagate_to(sep_var)) {
    Domain pdom=sep_var->get_dom(); // propagated domain
	
    vector<int> gm; // the "materialized function" g^m (from the paper)
    gm.resize(sep_var->get_ubound()+1);

    MaterializedFunction *g_child =
      (MaterializedFunction *)
      funcs_[funcs_.size()-1]; // the message from the child
    
    if (VERBOSE) cout << "Message of child: "<<g_child->get_name()<<endl;
	
    gm[0]=INT_MIN;
    gm[1]=g_child->value(0);
	
    for(int j=2; j <= sep_var->get_ubound(); j++) {
      gm[j] = max(infty_add(gm[j-1],sc_->gamma),
		  g_child->value(j-1));
    }
	
    for (int j=pdom.first(); j!=-1; j=pdom.next(j)) {
      sep_var->set_singleton(j); // set domain to {j}
      f->set_value(max((j>0)?infty_add(gm[j], sc_->sigma(i_,j)):INT_MIN,
		       infty_add(g_child->value(j), sc_->gamma)));
    }
  }
  sep_var->set_dom(dom); // restore old domain of enumerated variable

  return f;
}




/************************************************************
 * Special Problem Instance of Pure Sequence Alignment
 *
 */

var_ptrs_t x;
func_ptrs_t f;
func_ptrs_t leq;

vector<AligCluster *> clusters;

CTD alignmentCTD;


void init_alignment_problem(int n, int m, const ScoringInfo *myScoringInfo) {
  for (int i=0; i<=n+1; i++) {
    Variable *v=new Variable("x_"+itoa(i),0,m);
    x.push_back(v);
  }
  x[0]->set_singleton(0);
  x[n+1]->set_singleton(m+1);

  f.resize(n+2);
  leq.resize(n+2);

  for (int i=1; i<=n+1; i++) {
    f[i]=new ScoreFunction(myScoringInfo,i,x[i-1],x[i]);
    leq[i]=new LeqConstraint(i,x[i-1],x[i]);
  }
    
  // the variables x together with the functions f and leq describe
  // our weighted CSP
}

void init_alignment_ctd(int n, int m, const ScoringInfo *myScoringInfo) {
  clusters.resize(n+2);
  clusters[0]=0;

  Cluster *prev_c=0;

  for (int i=n+1; i>0; i--) {
    AligCluster *c;
    c=new AligCluster("Cluster_"+itoa(i),i,myScoringInfo);
    if (i==1 || i==n+1) c->marginalization_method(AligCluster::GENERAL_MARG);
    c->register_var(x[i-1]);
    c->register_var(x[i]);
    c->register_func(f[i]);
    c->register_cons(leq[i]);

    if (prev_c)
      prev_c->register_child(c); // this also sets parent pointer in c
    else
      alignmentCTD.set_root(c);

    clusters[i]=c;

    prev_c=c;
  }
}

/* extract region with uppercase characters out of string */
void get_segment(const string &seq, int *start, int *end) {
  *start=-1;
  *end=-1;
  int i;
    
  for (i=0; i<(int)seq.size() && islower(seq[i]); i++)
    ;
  *start=i+1;
    
  for (i=seq.size()-1; i>=0 && islower(seq[i]); i--)
    ;
  *end=i+1;
    
  return;
}

/* extract regions with uppercase characters out of string
 and put starts and end in a list of ints */
void extract_segments(const string &seq,
		      vector<int> &starts, 
		      vector<int> &ends) {
  enum {OUT,IN};
  int state=OUT;
  for (uint i=0; i<seq.size(); i++) {
    if (state==OUT && isupper(seq[i])) {
      starts.push_back(i+1);
      state=IN;
    } else if (state==IN && islower(seq[i])) {
      ends.push_back(i);
      state=OUT;
    }
  }
  if (state==IN) ends.push_back(seq.size());

  return;
}


class CountConstraint : public Function {
  const int l_;
  const int l_pr_;
public:
  CountConstraint(int i, Variable *a0, Variable *a1, 
		  Variable *x0, Variable *x1,
		  int l, int l_pr)
    : Function("c_"+itoa(i)), l_(l),l_pr_(l_pr)
  {
    vars_.push_back(a0);
    vars_.push_back(a1);
    vars_.push_back(x0);
    vars_.push_back(x1);
  }

  virtual ~CountConstraint();

  bool propagate_to(Variable *v) const {

    if (v->get_ubound()<v->get_lbound()) return false;
	
    Variable *a0=vars_[0];
    Variable *a1=vars_[1];
    Variable *x0=vars_[2];
    Variable *x1=vars_[3];
	
    if (v==a0 || v==a1) {
      // ist x0<x1 and l<=x1<=l_pr entailed (sonst keine Propagation)
      int add;
	    
      if ( x0->get_ubound() < x1->get_lbound()
	   && l_ <= x1->get_lbound() && x1->get_ubound() <= l_pr_) {
	add = 1;
      } else if (x0->get_lbound() >= x1->get_ubound()
		 || l_ > x1->get_ubound() || x1->get_lbound() > l_pr_) {
	add=0;
      } else {
	if (VERBOSE) {
	  //cout << "Delayed"<<"["<<l_<<" "<<l_pr_<<endl;
	  //x0->print_debug();x1->print_debug();
	  //cout<<"]"<<endl;
	}
	return true;
      }
      // cout << "Add: "<<add<<endl;
	    
      if (v==a0)
	{
	  return
	    a0->set_geq(a1->get_lbound()-add)
	    && 
	    a0->set_leq(a1->get_ubound()-add);
	} 
      else if (v==a1)
	{
	  return
	    a1->set_geq(a0->get_lbound()+add)
	    && 
	    a1->set_leq(a0->get_ubound()+add);    
	}
    }

    if (a1->get_lbound()>a0->get_ubound()) {
      if (v==x0) {
	return x0->set_leq(x1->get_ubound()-1);		
      } else if (v==x1) {
	return x1->set_geq(x0->get_lbound()+1)
	  && x1->set_geq(l_)
	  && x1->set_leq(l_pr_);
      }
    }
	
    return true;
  }

  int value() const {
    assert(0/* this should not be used*/);
  }
    
};
 
CountConstraint::~CountConstraint() {}


void add_segment_constraint(int startA, int endA, int startB, int endB, 
			    float percent, const ScoringInfo *myScoringInfo) {
  var_ptrs_t a;
  func_ptrs_t count_cons;
     
  a.resize(endA+1);
  count_cons.resize(endA+1);
    
  for (int i=startA-1;i<=endA; i++) {
    a[i]=new Variable("a_"+itoa(i),
		      max(0,(int)ceil(percent*(endA-startA+1))-(endA-i)),
		      i-startA+1);
    if (VERBOSE) a[i]->print_debug();
  }
        
  for (int i=startA; i<=endA; i++) {
    count_cons[i]=new CountConstraint(i,a[i-1],a[i],x[i-1],x[i],startB,endB);
  }
  
  for (int i=startA; i<=endA; i++) {
    AligCluster *c=clusters[i];
    c->register_var(a[i-1]);
    c->register_var(a[i]);
    c->register_cons(count_cons[i]);  
    
    c->marginalization_method(AligCluster::GENERAL_MARG);
    
    /* welche cluster können nicht mehr mit der optimierten Methode
       marginalisiert werden ????? */
  }
}


void add_precedence_constraint(int i, int j) {
  x[i]->set_leq(j); // x_i <= j
  for (unsigned int k=i-1; k>0; --k)
      x[k]->set_leq(j-1); // x_{k < j  
  clusters[i+1]->marginalization_method(AligCluster::GENERAL_MARG);
}


//  X_i<j     ==>  X_i==X_i-1
//
class NoMatchConstraintI: public Function {
    int j_;
  
public:
  NoMatchConstraintI(int i, int j, Variable *x_i_minus_1, Variable *x_i)
  : Function("NoMatchI "+itoa(i)+"-"+itoa(j)), j_(j) {
    vars_.push_back(x_i_minus_1);
    vars_.push_back(x_i);
  }
  
  virtual bool propagate_to(Variable *v) const {
    if (v==vars_[0]) { // X_i<j ==>  X_i==X_i-1
	if (vars_[1]->get_ubound()<j_) {  // if X_i<j
	    return vars_[0]->restrict_dom(vars_[1]->get_dom());   // x_{i-1} == x_{i}
      }
    } else if (v==vars_[1]) {
	// X_i<j ==>  X_i==X_i-1
	
	if (vars_[1]->get_ubound()<j_) {
	    return vars_[1]->restrict_dom(vars_[0]->get_dom());
	}
    }
    return true;
  }
  int value() const {
    assert(false/*never used*/);
  }

};


//  X_i+1==j     ==>  X_i==j
//
class NoMatchConstraintII: public Function {
    int j_;
  
public:
  NoMatchConstraintII(int i, int j, Variable *x_i, Variable *x_i_plus_1)
  : Function("NoMatchII "+itoa(i)+"-"+itoa(j)), j_(j) {
    vars_.push_back(x_i);
    vars_.push_back(x_i_plus_1);
  }
    
  virtual bool propagate_to(Variable *v) const {
      if (v==vars_[0]) {
	  if (vars_[1]->is_singleton()
	      && vars_[1]->get_singleton()==j_) { // if x_{i+1}==j
	      return vars_[0]->restrict_dom(Domain(j_,j_));       // then x_i=j;
	  }
      } else if (v==vars_[1]) { // X_i!=j  ==>  X_{i+1}!=j
	  if (vars_[0]->get_ubound() < j_ 
	      || vars_[0]->get_lbound() > j_ ) { // if X_i != j
	      return vars_[1]->set_geq(j_+1);    // then X_{i+1} > j_ (X_{i+1}>=j_ is known)
	  } 
    }
      return true;
  }
  int value() const {
    assert(false/*never used*/);
  }

};




// i and j are not matched
// every position before i can only match to something <j
// every position after  i can only match to something >j
//
void add_anchor_constraint( unsigned int i, unsigned int j ) {
  
    // the anchor constraint is equivalent to
    // X_{i-1}<j && X_i<=j && X_{i+1}>=j
    // && X_i<j     ==>  X_i==X_i-1
    // && X_i+1==j  ==>  X_i==j

  assert(i>=1);
  assert(j>=1);
  assert(i<=x.size());
  
  x[i-1]->set_leq(j-1);  // X_{i-1} <  j
  x[i]  ->set_leq(j);    // X_i     <= j
  x[i+1]->set_geq(j);    // X_{i+1} >= j
  
  // no match with j at position i+1
  
  clusters[i]  ->register_cons(new NoMatchConstraintI (i,j, x[i-1], x[i]   ));
  clusters[i+1]->register_cons(new NoMatchConstraintII(i,j, x[i],   x[i+1] ));
  
  clusters[i]->marginalization_method(AligCluster::GENERAL_MARG);
  clusters[i+1]->marginalization_method(AligCluster::GENERAL_MARG);
  clusters[i+2]->marginalization_method(AligCluster::GENERAL_MARG);
}



/*
  write alignment from assignment of x variables (typically called
  after backtrace)
  
  PRE: variables x are singleton
*/
void write_alignment(const string &seqA, const string &seqB) {
    
  string aliA="";
  string aliB="";
    
  uint j=1; // j points to the last unused element in b

  for (uint i=1; i<=seqA.length(); i++) {
    if ( x[i]->get_singleton() > x[i-1]->get_singleton() ) {
      for (; j < (uint)x[i]->get_singleton(); j++) {
	aliA += "-";
	aliB += seqB[j-1];
      }
      aliA+=seqA[i-1];
      aliB+=seqB[j-1];
      j++;
    } else {
      aliA += seqA[i-1];
      aliB += "-";
    }
  }
    
  for (; j <= seqB.length(); j++) {
    aliA += "-";
    aliB += seqB[j-1];
  }
    

  if (VERBOSE) cout << "BACKTRACE: "<<"SeqA ";
  cout <<aliA<<endl;
  if (VERBOSE) cout << "BACKTRACE: "<<"SeqB ";
  cout <<aliB<<endl;   
}


int main(int argc, char **argv) {

  int gap_cost=-7;
  //int match_cost=3;
  //int mismatch_cost=-1;
  string simmatrixfile="Matrices/PAM250";
  //string simmatrixfile="Matrices/NUCL";
  
  if (argc!=1) {
      cerr << "USAGE: "<<argv[0]<<endl; 
      exit(-1);
  }
  
  istream &in=std::cin;
  if (!in) {
    cerr << "Cannot read file "<<argv[1]<<endl;
    exit(-1);
  }
  
  char line[BUF_SIZE];
  
  string seqA,nameA;
  string seqB,nameB;

  vector<int> startsA,endsA,startsB,endsB;
  bool segments_extracted=false;

  int n;
  int m;
        
  ScoringInfo *myScoringInfo=0;
  
  enum {NO_SEQ,SEQ_A,SEQ_B,SEQS_ALIGNED};
  int mode=NO_SEQ;

  while(in.getline(line,BUF_SIZE)) {
    istringstream linestream(line);
    
    string first;
    linestream >> first;
    
    if (first[0]=='#') {
      // cout << "COMMENT: "<<line<<endl;
    } else if(first[0]=='>') {
      if (mode==NO_SEQ) {
	linestream >> nameA;
	mode=SEQ_A;
      }else if (mode==SEQ_A) {
        linestream >> nameB;
	mode=SEQ_B;
      } else {
	cerr << "Only two input sequence can be given."<<endl; exit(-1);
      }
    } else if (first[0]=='-') {
	if (myScoringInfo!=0) {
	    cerr << "Scoring options have to be specified before constraints."<<endl
		 << "Option \""<<line<<"\" is ignored."<<endl;
	} else {
	    if (first=="-matrix") {
		linestream >> simmatrixfile;
	    } else if (first=="-gap") {
		linestream >> gap_cost;
	    }
	    //if (VERBOSE)
	    cout <<line<<endl;
	}
    }
    else if (first[0]=='!') {
      if (mode!=SEQS_ALIGNED) {
	if (nameA=="" || nameB=="") { 
	  cout << "You need to specify two sequences before "
	       << "stating constraints."<<endl;
	  exit(-1);
	}
	
	/*
	cout << "SeqA ("<<nameA<<") "<<seqA<<endl;
	cout << "SeqB ("<<nameB<<") "<<seqB<<endl;
	*/
	
	n=seqA.size();
	m=seqB.size();
        
	// myScoringInfo = new ScoringInfo(seqA,seqB,gap_cost,match_cost,mismatch_cost);
	myScoringInfo = new ScoringInfo(seqA,seqB,gap_cost,simmatrixfile);

	init_alignment_problem(n,m,myScoringInfo);
	init_alignment_ctd(n,m,myScoringInfo);
	
	mode=SEQS_ALIGNED;
      }

      if (first=="!segment") {
	if (!segments_extracted) {
	  
	  extract_segments(seqA,startsA,endsA);
	  extract_segments(seqB,startsB,endsB);
	  
	  segments_extracted=true;
	
	}
	
	double percent;	
	
	string stringA;
	string stringB;
	uint intA;
	uint intB;

	linestream >> stringA >> intA >> stringB >> intB >> percent;
	
	uint startA, endA, startB, endB;

	if (stringA == "seg") {
	  uint segmentA=intA;
	  if (segmentA<1 || segmentA > startsA.size()) {
	    cerr << "There is no segment "<<segmentA<<" in seqA."<<endl;
	    exit(-1);
	  }
	  startA=startsA[segmentA-1];
	  endA=endsA[segmentA-1];
	} else {
	  startA=atoi(stringA.c_str());
	  endA=intA;
	  
	  if (startA<1 || startA > endA || endA>seqA.size()) {
	    cerr << "Invalid segment ("<<startA<<"-"<<endA<<")."<<endl;
	    exit(-1);
	  }
	}
	
	if (stringB == "seg") {	
	  uint segmentB=intB;
	  if (segmentB<1 || segmentB > startsB.size()) {
	    cerr << "There is no segment "<<segmentB<<" in seqB."<<endl;
	    exit(-1);
	  }
	  startB=startsB[segmentB-1];
	  endB=endsB[segmentB-1];
	} else {
	  startB=atoi(stringB.c_str());
	  endB=intB;
	  if (startB<1 || startB > endB || endB>seqB.size()) {
	    cerr << "Invalid segment ("<<startB<<"-"<<endB<<")."<<endl;
	    exit(-1);
	  }
	}
	
	if (VERBOSE) cout << "Segment ("<<startA<<"-"<<endA<<") of seqA matchs to segment ("
	     <<startB<<"-"<<endB<<") of seqB by at least "
	     <<100*percent<<"%"<<endl; 
	
	add_segment_constraint( startA, endA, startB, endB,
				percent, myScoringInfo );
	
      } else if ( first == "!precedence") {
	int i; 
	int j;
	linestream >>  i >> j;
	add_precedence_constraint( i, j );
	if (VERBOSE) cout << "a_"<<i<<" precedes b_"<<j<<endl;
      } else if ( first == "!anchor" ) {
	int i; 
	int j;
	linestream >>  i >> j;
	add_anchor_constraint( i, j );
	if (VERBOSE) cout << "a_"<<i<<" anchored to b_"<<j<<endl;
      } else {
	cerr << "Unknown constraint (ignored): "<<first<<endl;
      }
    } else if (mode==SEQ_A) {
      seqA+=line;
    } else if (mode==SEQ_B) {
      seqB+=line;
    } else if (first=="") {
      // ignore empty line
    } else {
      cerr << "Warning: ignored line \""<<line<<"\""<<endl;
    }
  }
  
  
  if (mode!=SEQS_ALIGNED) { // this is only for the case of
			    // unconstrained alignment
    if (nameA=="" || nameB=="") { 
      cerr << "You need to specify two sequences."<<endl;
      exit(-1);
    }
    
    /*
    cout << "SeqA ("<<nameA<<") "<<seqA<<endl;
    cout << "SeqB ("<<nameB<<") "<<seqB<<endl;
    */

    n=seqA.size();
    m=seqB.size();

    //myScoringInfo = new ScoringInfo(seqA,seqB,gap_cost,match_cost,
    // mismatch_cost);   
    myScoringInfo = new ScoringInfo(seqA,seqB,gap_cost,simmatrixfile);

    init_alignment_problem(n,m,myScoringInfo);
    init_alignment_ctd(n,m,myScoringInfo);
    
  }

  if (VERBOSE) alignmentCTD.print_debug();
    
  // get root cluster
  Cluster *root=alignmentCTD.get_root();

  MaterializedFunction * result = 
    root->marginalize(); //  marginalize to the empty set
  // result is a constant function and contains the final result of
  // the CTE algorithms

    
  if (VERBOSE) 
      cout <<endl<<endl
	   <<"------------------------------------------------------------"
	   <<endl<<"Backtrace"<<endl;
  
  if (result->value()>INT_MIN) {
    root->backtrace();
    
    write_alignment(seqA,seqB);
  } else {
      cerr << "No alignment satisfies the constraints."<<endl;
  }
  cout <<result->value()<<endl;

  for (var_ptrs_t::iterator it=x.begin(); x.end() != it; ++it) delete *it;
  alignmentCTD.free();
    
  exit(0);
}
