#ifndef LEQ_CONSTRAINT_HH
#define LEQ_CONSTRAINT_HH

#include "variable.hh"
#include "functions.hh"

/*
  Constraint "v0 <= v1"
*/
class LeqConstraint: public Function {
public:
    LeqConstraint(int i,Variable *v0, Variable *v1): Function("leq_"+itoa(i)) {
	vars_.push_back(v0);
	vars_.push_back(v1);
    }

    virtual ~LeqConstraint();
    
    bool propagate_to(Variable *v) const {
	if (v==vars_[0]) {
	    return
		vars_[0]->set_ubound(
		    std::min(
			vars_[0]->get_ubound(),
			vars_[1]->get_ubound()
			)
		    );
 	} else if (v==vars_[1]) {
	    return
		vars_[1]->set_lbound(
		    std::max(
			vars_[0]->get_lbound(),
			vars_[1]->get_lbound()
			)
		    );
	} else {
	    return true; // constraint has no effect on this variable
	}
    }
    
    int value() const {
	return 
	    (vars_[0]->get_singleton()<=vars_[1]->get_singleton())
	    ?0
	    :INT_MIN;
    }
};


#endif // LEQ_CONSTRAINT_HH
